

import calendar
import os
import random

from matplotlib.dates import DateFormatter
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
import pandas

import c2sm

from smartcarb import misc
from smartcarb import map
from smartcarb import config as cfg


import dplume.io


DLON_ORBIT = 24.146341463414636

COLORS = [
    '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2',
    '#7f7f7f', '#bcbd22', '#17becf'
]



def visualize(data, do_zoom=True, show_clouds=True, use_noise=True,
              is_sentinel_5=False, data7=None):

    domain = cfg.DOMAINS['smartcarb']
    lon_o, lat_o, radius = data.attrs['plume location (lon, lat, radius)']

    # tracer field
    if not use_noise:
        noise = 0.0

    clouds = data.is_cloudy if show_clouds else None

    if data.tracer == 'CO2':
        mean_bg = np.nanmean(data.CO2)
        fig = c2sm.vis.make_level2_map(
            data.lonc, data.latc, data.CO2 - mean_bg, vmin=-2, vmax=2,
            label='XCO$_2$ signal (ppm)', clct=clouds, domain=domain
        )
    else:
        vmax = 10 if is_sentinel_5 else 20
        fig = c2sm.vis.make_level2_map(
            data.lonc, data.latc, data.NO2 / 1e15, vmin=0, vmax=vmax, truncate_cmap=True,
            label='NO$_2$ column (10$^{15}$ molecules per cm$^2$)', clct=clouds,
            domain=domain
        )
    #
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)

    # lines and labels
    lines = ax.plot(cfg.pollon, cfg.pollat, marker='s', mec='k', mfc='w', ls='',
            transform=ccrs.PlateCarree())
    labels = ['cloud fraction > %d%%' % (100*clouds.threshold)]

    # true plumes
    for tracer, ls in [('CO2', '-'), ('NO2', '--')]:
        name = '%s_true_plume' % tracer
        label = '%s plume' % tracer.replace('2', '$_2$')

        if hasattr(data, name):
            lines += c2sm.vis.draw_contours(ax, data.lon, data.lat,
                                            data[name], label=label, ls=ls)
            labels += [label]
            continue

        if hasattr(data7, name):
            lines += c2sm.vis.draw_contours(ax, data7.lon, data7.lat,
                                            data7[name], label=label, ls=ls)
            labels += ['%s (Sentinel-7)' % label]

    # detected plumes
    if data.detected_plume is not None and np.sum(data.detected_plume) > 0:

        lon0 = data.lon.values[data.detected_plume]
        lat0 = data.lat.values[data.detected_plume]

        lines += ax.plot(lon0, lat0, marker='o',
                mec='none', ms=1.5, color='k', ls='',
                transform=ccrs.PlateCarree()
        )
        if hasattr(data, 'confusion_matrix_30'):
            extra = ' (PPV = %.2f)' % data.confusion_matrix_30.PPV
        else:
            extra = ''

        q = data.attrs['probability for z-value']
        ns = data.attrs['size of neighborhood']

        labels += [
            'detected plume (q = %.3f, n = %d)\n' % (q, ns) +
            'with %d pixels%s' % (np.sum(data.detected_plume), extra)
        ]

    if do_zoom:
        try:
            rlon_0, rlat_0 = c2sm.transform.rotpole2wgs(
                data.CO2_true_plume.center_lon,
                data.CO2_true_plume.center_lat,
                cfg.pollon, cfg.pollat, inverse=True)
        except AttributeError:
            rlon_0, rlat_0 = c2sm.transform.rotpole2wgs(lon_o, lat_o, cfg.pollon,
                                              cfg.pollat, inverse=True)

        dy = 1.1
        dx = dy * 1.12
        xmin, xmax = rlon_0 - dx, rlon_0 + dx
        ymin, ymax = rlat_0 - dy, rlat_0 + dy

        try:
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
        except ValueError:
            pass

    if lines:
        ax.legend(lines, labels, loc="upper left", numpoints=1, markerscale=1.5,
                fontsize='small')


    return fig




def detection_limits(filename):
    d = pandas.read_csv(filename, index_col=0)

    fig = plt.figure()
    ax = fig.add_subplot(111)


    valids = np.all(np.isfinite(d.values), axis=1)
    limits = d.values[valids]

    ax.bar(np.arange(5), np.mean(limits, 0), yerr=np.std(limits, 0))

    ax.set_ylim(0,0.6)
    ax.set_ylabel('detection limit (ppm)')
    ax.grid(axis='y')

    ax.set_xticks(np.arange(5))
    ax.set_xticklabels([c.replace('2_', '$_2$\n') for c in d.columns])

    return ax



def month_labels(ax=None):
    if ax is None:
        ax = plt.gca()

    numbers = np.arange(1,13)
    labels = [calendar.month_abbr[n] for n in numbers]
    ax.set_xticks(numbers)
    ax.set_xticklabels(labels)



def date_labels(ax=None, fmt='%b', axis='x'):

    fmt = DateFormatter(fmt)

    if 'x' in axis:
        ax.xaxis.set_major_formatter(fmt)

    if 'y' in axis:
        ax.yaxis.set_major_formatter(fmt)


def plot_obs_per_month_and_scenario(nmax=99, nsat=3, cc=1, show=True):

    products = [
        ('CO2', 0.5, 37),
        ('CO2', 0.7, 37),
        ('CO2', 1.0, 37),
        ('NO2', 'S5', 1),
        ('NO2', 'low', 37),
        ('NO2', 'high', 37),
    ]

    product2label = {
        ('CO2', 1.0): 'CO$_2$, 1.0 ppm',
        ('CO2', 0.7): 'CO$_2$, 0.7 ppm',
        ('CO2', 0.5): 'CO$_2$, 0.5 ppm',
        ('NO2', 'S5'):   'NO$_2$, Sentinel-5',
        ('NO2', 'low'):  r'NO$_2$, 1$\times$10$^{15}$ cm$^{-2}$',
        ('NO2', 'high'): r'NO$_2$, 2$\times$10$^{15}$ cm$^{-2}$'
    }

    if show:
        fig = plt.figure(figsize=[11, 4])
        ax = fig.add_subplot(111)

    lon_eqs = np.linspace(0, DLON_ORBIT, nsat+1)[:-1]

    values = []
    errors = []

    for i,(tracer, ssp, ns) in enumerate(products):

        # estimate uncertainty
        data = dplume.io.read_results(tracer, ssp, ns, nsat=None)
        sample = [
            sum(data['TP%d' % cc][(data.PPV30 >= 0.8) & (data.lon_eq == l)] >= 100)
            for l in sorted(set(data.lon_eq))
        ]
        error = np.std([sum(random.sample(sample, nsat)) for _ in range(5000)])
        errors.append(error)

        data = dplume.io.read_results(tracer, ssp, ns, nsat=nsat)
        months, days = days_per_month(data, lon_eqs, nmax, cc=cc)
        cdays = days.sum(0)

        if show:
            ax.bar(months+i/7.0-0.5+1/7, cdays, width=1/7,
                   label='%s (%d)' % (product2label[tracer,ssp], sum(cdays)))

        values.append(sum(cdays))



    if show:
        month_labels(ax)

        ax.set_xlabel('months in 2015')
        ax.set_ylabel('number of plume observations')

        plt.legend(ncol=3, framealpha=0.4, title='Instrument scenario (number of plumes)')
        plt.tight_layout()
        plt.grid(axis='y')
        ax.set_ylim(0,9)

    # add black line showing the max. number of detectable plumes
    data = dplume.io.read_overpasses('Berlin', nsat=nsat)
    months, days = days_per_month(data, lon_eqs, nmax, cc=cc)
    cdays = days.sum(0)

    values.append(sum(cdays))

    data = dplume.io.read_overpasses('Berlin', nsat=None)
    sample = [
        sum(data['pixels (<1%)'][data.lon_eq == l] >= 100)
        for l in sorted(set(data.lon_eq))
    ]
    error = np.std([sum(random.sample(sample, nsat)) for _ in range(5000)])
    errors.append(error)

    if show:
        for i,month in enumerate(months):
            ax.plot([month-0.43, month+0.43], [cdays[i], cdays[i]], 'k-')


    return values,errors




def plot_plumes_per_sat(days=None):
    sats = np.arange(1,7)

    labels = [
        r'CO$_2$, 0.5 ppm',
        r'CO$_2$, 0.7 ppm',
        r'CO$_2$, 1.0 ppm',
        r'NO$_2$, Sentinel-5',
        r'NO$_2$, 1$\times$10$^{15}$ cm$^{-2}$',
        r'NO$_2$, 2$\times$10$^{15}$ cm$^{-2}$',
        'real plumes'
    ]

    if days is None:
        days = np.array([
            plot_obs_per_month_and_scenario(nmax=99, nsat=i, cc=1, show=False)
            for i in sats
        ])

    fig = plt.figure()
    ax = fig.add_subplot(111)

    for i in range(7):
        if i == 6:
            color = 'k'
        else:
            color = None

        line = ax.errorbar(sats, days[:,0,i], days[:,1,i], marker='o',
                           color=color, ls='', label=labels[i])[0]

        res = scipy.stats.linregress(sats, days[:,0,i])
        print(i, res)

        x = np.linspace(0.8,6.2)
        ax.plot(x, res.slope * x + res.intercept, '-', color=line.get_color())

    ax.set_xlim(0.6, 6.4)
    ax.set_ylim(0,70)

    ax.set_xlabel('number of satellites')
    ax.set_ylabel('number of plumes per year')

    plt.legend()







def plot_obs_per_month_and_nsat(filename, nmax=99, index_col=0,
                                only_overlaps=False, cloud_thr=0.01):
    """
    Plots for different number of satellites, the number of plume observations
    per month with more than nmax XCO2 pixels (cloud fraction < 1%).
    """
    data = pandas.read_csv(filename, index_col=index_col, sep='\t',
                           parse_dates=True, comment='#')

    fig = plt.figure(figsize=[11, 4])
    ax = fig.add_subplot(111)

    for i,nsat in enumerate(range(1,7)):

        months, days = days_per_month(data, nsat, nmax, cloud_thr=cloud_thr,
                                      only_overlaps=only_overlaps)

        ax.bar(months+i/7.0-0.5+1/7, days, width=1/7,
               label='%d (%d)' % (nsat, sum(days)))

    month_labels(ax)

    ax.set_xlabel('months in 2015')
    ax.set_ylabel('number of plume observations')
    ax.set_ylim(0,9)

    plt.legend(ncol=3, title='number of satellites (number of plumes)')
    plt.tight_layout()
    plt.grid(axis='y')



def days_per_month(data, nsat, nmax=99, scaling=None, cloud_thr=0.01,
                   only_overlaps=False):

    data = misc.filter_dataset(data, nsat)

    if 'pixels (<1%)' in data.columns:
        pixels = data['pixels (<%d%%)' % (100 * cloud_thr)]

        if only_overlaps:
            print(pixels.shape, data.does_overlap.shape)
            pixels = pixels[data.does_overlap]
    else:
        pixels = data['TP%d' % (100 * cloud_thr)][data.PPV30 >= 0.80]

    if scaling is None:
        scaling = np.ones(13)

    days = np.array([
        scaling[m] * sum(pixels[pixels.index.month == m] > nmax)
        for m in range(1,13)
    ])

    return np.arange(1,13), np.array(days)




def plot_obs_per_month_boxplot(filename, nsat=4, nmax=-1, use_percent=False,
                               add_sub_constellations=False):

    # time, lon_eq, cloud-free pixels
    #data = pandas.read_csv(filename, index_col=0, parse_dates=True)
    data = filename

    lon_eqs = np.linspace(0, DLON_ORBIT, nsat+1)[:-1]

    if add_sub_constellations and nsat in [4,6]:
        lon_eqs = np.concatenate([lon_eqs[::2], lon_eqs[1::2]])

    if use_percent:
        mdays = np.array(calendar.mdays)
        mdays[12] = 25.0 # only 25 days have been simulated in Dec.
        scaling = 100.0 / mdays
    else:
        scaling = np.ones(13)

    months, days = days_per_month(data, nsat, nmax)

    fig = plt.figure()
    ax = fig.add_subplot(111)

    for i,lon_eq in enumerate(lon_eqs):
        label = '%s (%d°)' % ('ABCDEF'[i], round(lon_eq))
        bottom = np.sum(days[:i], axis=0)
        ax.bar(months, days[i], bottom=bottom, label=label)



    if add_sub_constellations and nsat in [4,6]:
        ax.set_ylim(0,10)
        for i in [1, 3 if nsat == 6 else 2, nsat]:
            pass
            #cdays = days[:i].sum(0)
            #ax.plot(months, cdays - 0.12, 'kv')4
            #for j,m in enumerate(months):
            #    ax.plot([m-0.4, m+0.4], [cdays[j], cdays[j]], 'k-', lw=2)



    month_labels(ax)

    ax.set_xlabel('months in 2015')
    ax.set_ylabel('number of plume observations')

    if add_sub_constellations and nsat in [4,6]:
        plt.legend(ncol=2, title='satellite')
    else:
        plt.legend(ncol=nsat, title='satellites')

    plt.tight_layout()

    return data












